"""
Contains constants variables to support CoDICE processing.

Notes
-----
The ``plan_id``, ``plan_step``, and ``view_id`` mentioned in this module are
derived from the packet data.

Some notable acronyms:
SW = SunWard
NSW = Non-SunWard
PUI = PickUp Ion
ESA = ElectroStatic Analyzer
"""

from typing import Any

import numpy as np

from imap_processing.codice.utils import CODICEAPID, CoDICECompression

# Grouping of APIDs used to signify similar L1a processing
APIDS_FOR_SCIENCE_PROCESSING = [
    CODICEAPID.COD_HI_INST_COUNTS_AGGREGATED,
    CODICEAPID.COD_HI_INST_COUNTS_PRIORITIES,
    CODICEAPID.COD_HI_INST_COUNTS_SINGLES,
    CODICEAPID.COD_HI_SECT_SPECIES_COUNTS,
    CODICEAPID.COD_LO_INST_COUNTS_AGGREGATED,
    CODICEAPID.COD_LO_INST_COUNTS_SINGLES,
    CODICEAPID.COD_LO_SW_ANGULAR_COUNTS,
    CODICEAPID.COD_LO_NSW_ANGULAR_COUNTS,
    CODICEAPID.COD_LO_SW_PRIORITY_COUNTS,
    CODICEAPID.COD_LO_NSW_PRIORITY_COUNTS,
    CODICEAPID.COD_LO_SW_SPECIES_COUNTS,
    CODICEAPID.COD_LO_NSW_SPECIES_COUNTS,
]

# Mapping between descriptors and APID
CODICEAPID_MAPPING = {
    "hskp": CODICEAPID.COD_NHK,
    "lo-ialirt": CODICEAPID.COD_LO_IAL,
    "lo-pha": CODICEAPID.COD_LO_PHA,
    "lo-sw-priority": CODICEAPID.COD_LO_SW_PRIORITY_COUNTS,
    "lo-sw-species": CODICEAPID.COD_LO_SW_SPECIES_COUNTS,
    "lo-nsw-species": CODICEAPID.COD_LO_NSW_SPECIES_COUNTS,
    "lo-sw-angular": CODICEAPID.COD_LO_SW_ANGULAR_COUNTS,
    "lo-nsw-angular": CODICEAPID.COD_LO_NSW_ANGULAR_COUNTS,
    "lo-nsw-priority": CODICEAPID.COD_LO_NSW_PRIORITY_COUNTS,
    "lo-counters-aggregated": CODICEAPID.COD_LO_INST_COUNTS_AGGREGATED,
    "lo-counters-singles": CODICEAPID.COD_LO_INST_COUNTS_SINGLES,
    "hi-ialirt": CODICEAPID.COD_HI_IAL,
    "hi-pha": CODICEAPID.COD_HI_PHA,
    "hi-counters-aggregated": CODICEAPID.COD_HI_INST_COUNTS_AGGREGATED,
    "hi-counters-singles": CODICEAPID.COD_HI_INST_COUNTS_SINGLES,
    "hi-omni": CODICEAPID.COD_HI_OMNI_SPECIES_COUNTS,
    "hi-sectored": CODICEAPID.COD_HI_SECT_SPECIES_COUNTS,
    "hi-priority": CODICEAPID.COD_HI_INST_COUNTS_PRIORITIES,
}

# Numerical constants
SPIN_PERIOD_CONVERSION = 0.00032

# CDF variable names used for lo data products
LO_COUNTERS_SINGLES_VARIABLE_NAMES = ["apd_singles"]
LO_SW_ANGULAR_VARIABLE_NAMES = ["hplus", "heplusplus", "oplus6", "fe_loq"]
LO_NSW_ANGULAR_VARIABLE_NAMES = ["heplusplus"]
LO_SW_PRIORITY_VARIABLE_NAMES = [
    "p0_tcrs",
    "p1_hplus",
    "p2_heplusplus",
    "p3_heavies",
    "p4_dcrs",
]
LO_NSW_PRIORITY_VARIABLE_NAMES = ["p5_heavies", "p6_hplus_heplusplus"]
LO_SW_SPECIES_VARIABLE_NAMES = [
    "hplus",
    "heplusplus",
    "cplus4",
    "cplus5",
    "cplus6",
    "oplus5",
    "oplus6",
    "oplus7",
    "oplus8",
    "ne",
    "mg",
    "si",
    "fe_loq",
    "fe_hiq",
    "heplus",
    "cnoplus",
]
LO_NSW_SPECIES_VARIABLE_NAMES = [
    "hplus",
    "heplusplus",
    "c",
    "o",
    "ne_si_mg",
    "fe",
    "heplus",
    "cnoplus",
]
LO_IALIRT_VARIABLE_NAMES = [
    "heplusplus",
    "cplus5",
    "cplus6",
    "oplus6",
    "oplus7",
    "oplus8",
    "mg",
    "fe_loq",
    "fe_hiq",
]

# CDF variable names used for hi data products
HI_COUNTERS_SINGLES_VARIABLE_NAMES = ["tcr", "ssdo", "stssd"]
HI_OMNI_VARIABLE_NAMES = ["h", "he3", "he4", "c", "o", "ne_mg_si", "fe", "uh", "junk"]
HI_PRIORITY_VARIABLE_NAMES = [
    "Priority0",
    "Priority1",
    "Priority2",
    "Priority3",
    "Priority4",
    "Priority5",
]
HI_SECTORED_VARIABLE_NAMES = ["h", "he3he4", "cno", "fe"]
HI_IALIRT_VARIABLE_NAMES = ["h"]

# CDF variable names used for direct event data products
HI_PHA_CDF_FIELDS = [
    "NumEvents",
    "DataQuality",
    "SSDEnergy",
    "TOF",
    "SSD_ID",
    "ERGE",
    "MultiFlag",
    "Type",
    "SpinAngle",
    "SpinNumber",
]
HI_PHA_VARIABLE_NAMES = [
    f"P{n}_{field}" for n in range(6) for field in HI_PHA_CDF_FIELDS
]
LO_PHA_CDF_FIELDS = [
    "NumEvents",
    "DataQuality",
    "APDGain",
    "APD_ID",
    "APDEnergy",
    "TOF",
    "MultiFlag",
    "PHAType",
    "SpinAngle",
    "EnergyStep",
]
LO_PHA_VARIABLE_NAMES = [
    f"P{n}_{field}" for n in range(8) for field in LO_PHA_CDF_FIELDS
]

# lo- and hi-counters-aggregated data product variables are dynamically
# determined based on the number of active counters
# TODO: Try to convince Joey to move to lower case variable names with
#       underscores?
LO_COUNTERS_AGGREGATED_ACTIVE_VARIABLES = {
    "TCR": True,
    "DCR": True,
    "TOFPlusAPD": False,
    "TOFOnly": False,
    "PositionPlusAPD": False,
    "PositionOnly": False,
    "STAorSTBPlusAPD": False,
    "STAorSTBOnly": False,
    "Reserved1": False,
    "Reserved2": False,
    "SPOnly": False,
    "APDOnly": False,
    "LowTOFCutoff": False,
    "STA": True,
    "STB": True,
    "SP": True,
    "TotalPositionCount": True,
    "InvalidPositionCount": False,
    "ASIC1FlagInvalid": False,
    "ASIC2FlagInvalid": False,
    "ASIC1ChannelInvalid": False,
    "ASIC2ChannelInvalid": False,
    "TEC4TimeoutTOFNoPos": False,
    "TEC4TimeoutPosNoTOF": False,
    "TEC4TimeoutNoPosTOF": False,
    "TEC5TimeoutTOFNoPos": False,
    "TEC5TimeoutPosNoTOF": False,
    "TEC5TimeoutNoPosTOF": False,
}
LO_COUNTERS_AGGREGATED_VARIABLE_NAMES = [
    name
    for name, is_active in LO_COUNTERS_AGGREGATED_ACTIVE_VARIABLES.items()
    if is_active
]
HI_COUNTERS_AGGREGATED_ACTIVE_VARIABLES = {
    "DCR": True,
    "STO": True,
    "SPO": True,
    "Reserved1": False,
    "MST": True,
    "Reserved2": False,
    "Reserved3": False,
    "Reserved4": False,
    "Reserved5": False,
    "LowTOFCutoff": True,
    "Reserved6": True,
    "Reserved7": True,
    "ASIC1FlagInvalid": True,
    "ASIC2FlagInvalid": True,
    "ASIC1ChannelInvalid": True,
    "ASIC2ChannelInvalid": True,
}
HI_COUNTERS_AGGREGATED_VARIABLE_NAMES = [
    name
    for name, is_active in HI_COUNTERS_AGGREGATED_ACTIVE_VARIABLES.items()
    if is_active
]

# Energy tables for CoDICE-Hi data products. These values represent the edges
# of the bins, and are used in the CoDICE L1a pipeline to compute the centers
# and deltas of the bins, which then get stored in the CDF files for future use.
# These are defined in the "Data Products - Hi" tab of the "*-SCI-LUT-*.xml"
# spreadsheet that largely defines CoDICE processing.
IALIRT_ENERGY_TABLE = {
    "h": [
        0.05,
        0.070710678,
        0.1,
        0.141421356,
        0.2,
        0.282842712,
        0.4,
        0.565685425,
        0.8,
        1.13137085,
        1.6,
        2.2627417,
        3.2,
        4.5254834,
        6.4,
        9.050966799,
    ],
}

OMNI_ENERGY_TABLE = {
    "h": [
        0.05,
        0.070710678,
        0.1,
        0.141421356,
        0.2,
        0.282842712,
        0.4,
        0.565685425,
        0.8,
        1.13137085,
        1.6,
        2.2627417,
        3.2,
        4.5254834,
        6.4,
        9.050966799,
    ],
    "he3": [
        0.035355339,
        0.05,
        0.070710678,
        0.1,
        0.141421356,
        0.2,
        0.282842712,
        0.4,
        0.565685425,
        0.8,
        1.13137085,
        1.6,
        2.2627417,
        3.2,
        4.5254834,
        6.4,
    ],
    "he4": [
        0.035355339,
        0.05,
        0.070710678,
        0.1,
        0.141421356,
        0.2,
        0.282842712,
        0.4,
        0.565685425,
        0.8,
        1.13137085,
        1.6,
        2.2627417,
        3.2,
        4.5254834,
        6.4,
    ],
    "c": [
        0.025,
        0.035355339,
        0.05,
        0.070710678,
        0.1,
        0.141421356,
        0.2,
        0.282842712,
        0.4,
        0.565685425,
        0.8,
        1.13137085,
        1.6,
        2.2627417,
        3.2,
        4.5254834,
        6.4,
        9.050966799,
        12.8,
    ],
    "o": [
        0.025,
        0.035355339,
        0.05,
        0.070710678,
        0.1,
        0.141421356,
        0.2,
        0.282842712,
        0.4,
        0.565685425,
        0.8,
        1.13137085,
        1.6,
        2.2627417,
        3.2,
        4.5254834,
        6.4,
        9.050966799,
        12.8,
    ],
    "ne_mg_si": [
        0.01767767,
        0.025,
        0.035355339,
        0.05,
        0.070710678,
        0.1,
        0.141421356,
        0.2,
        0.282842712,
        0.4,
        0.565685425,
        0.8,
        1.13137085,
        1.6,
        2.2627417,
        3.2,
    ],
    "fe": [
        0.01767767,
        0.025,
        0.035355339,
        0.05,
        0.070710678,
        0.1,
        0.141421356,
        0.2,
        0.282842712,
        0.4,
        0.565685425,
        0.8,
        1.13137085,
        1.6,
        2.2627417,
        3.2,
        4.5254834,
        6.4,
        9.050966799,
    ],
    "uh": [0.01767767, 0.025, 0.035355339, 0.05, 0.070710678, 0.1],
    "junk": [0.05, 0.070710678],
}

SECTORED_ENERGY_TABLE = {
    "h": [0.05, 0.1, 0.2, 0.4, 0.8, 1.6, 3.2, 6.4, 12.8],
    "he3he4": [0.025, 0.05, 0.1, 0.2, 0.4, 0.8, 1.6, 3.2, 6.4],
    "cno": [0.025, 0.05, 0.1, 0.2, 0.4, 0.8, 1.6, 3.2, 6.4],
    "fe": [0.0125, 0.025, 0.05, 0.1, 0.2, 0.4, 0.8, 1.6, 3.2],
}

# TODO: Possibly move to consistent order of dimensions with other instruments
#       TBD after discussion with Joey and at the Science Team Meeting in Feb
# Various configurations to support processing of individual data products
# Much of these are described in the algorithm document in chapter 10 ("Data
# Level 1A")
DATA_PRODUCT_CONFIGURATIONS: dict[CODICEAPID | int, dict] = {
    CODICEAPID.COD_HI_IAL: {
        "dataset_name": "imap_codice_l1a_hi-ialirt",
        "energy_table": IALIRT_ENERGY_TABLE,
        "input_dims": {"esa_step": 15, "inst_az": 4},
        "instrument": "hi",
        "num_counters": 1,
        "num_spins": 4,
        "output_dims": {"esa_step": 15, "inst_az": 4},
        "support_variables": [
            "data_quality",
            "spin_period",
            "energy_h",
        ],
        "variable_names": HI_IALIRT_VARIABLE_NAMES,
    },
    CODICEAPID.COD_HI_INST_COUNTS_AGGREGATED: {
        "dataset_name": "imap_codice_l1a_hi-counters-aggregated",
        "input_dims": {},
        "instrument": "hi",
        "num_counters": len(
            HI_COUNTERS_AGGREGATED_VARIABLE_NAMES
        ),  # The number of counters depends on the number of active counters
        "output_dims": {},
        "support_variables": ["data_quality", "spin_period"],
        "variable_names": HI_COUNTERS_AGGREGATED_VARIABLE_NAMES,
    },
    CODICEAPID.COD_HI_INST_COUNTS_SINGLES: {
        "dataset_name": "imap_codice_l1a_hi-counters-singles",
        "input_dims": {
            "ssd_index": 12,
        },
        "instrument": "hi",
        "num_counters": 3,
        "output_dims": {
            "ssd_index": 12,
        },
        "support_variables": ["data_quality", "spin_period"],
        "variable_names": HI_COUNTERS_SINGLES_VARIABLE_NAMES,
    },
    CODICEAPID.COD_HI_INST_COUNTS_PRIORITIES: {
        "dataset_name": "imap_codice_l1a_hi-priority",
        "input_dims": {},
        "instrument": "hi",
        "num_counters": 6,
        "output_dims": {},
        "support_variables": ["data_quality", "spin_period"],
        "variable_names": HI_PRIORITY_VARIABLE_NAMES,
    },
    CODICEAPID.COD_HI_OMNI_SPECIES_COUNTS: {
        "dataset_name": "imap_codice_l1a_hi-omni",
        "energy_table": OMNI_ENERGY_TABLE,
        "input_dims": {"esa_step": 15, "inst_az": 4},
        "instrument": "hi",
        "num_counters": 8,
        "num_spins": 4,
        "output_dims": {"esa_step": 15, "inst_az": 4},
        "support_variables": [
            "data_quality",
            "spin_period",
            "energy_h",
            "energy_he3",
            "energy_he4",
            "energy_c",
            "energy_o",
            "energy_ne_mg_si",
            "energy_fe",
            "energy_uh",
            "energy_junk",
        ],
        "variable_names": HI_OMNI_VARIABLE_NAMES,
    },
    CODICEAPID.COD_HI_SECT_SPECIES_COUNTS: {
        "dataset_name": "imap_codice_l1a_hi-sectored",
        "energy_table": SECTORED_ENERGY_TABLE,
        "input_dims": {
            "esa_step": 8,
            "ssd_index": 12,
            "spin_sector_index": 12,
        },
        "instrument": "hi",
        "num_counters": 4,
        "num_spins": 16,
        "output_dims": {
            "esa_step": 8,
            "ssd_index": 12,
            "spin_sector_index": 12,
        },
        "support_variables": [
            "data_quality",
            "spin_period",
            "energy_h",
            "energy_he3he4",
            "energy_cno",
            "energy_fe",
        ],
        "variable_names": HI_SECTORED_VARIABLE_NAMES,
    },
    CODICEAPID.COD_LO_IAL: {
        "dataset_name": "imap_codice_l1a_lo-ialirt",
        "input_dims": {"spin_sector": 1, "esa_step": 128},
        "instrument": "lo",
        "num_counters": 9,
        "output_dims": {"spin_sector": 1, "esa_step": 128},
        "support_variables": [
            "energy_table",
            "acquisition_time_per_step",
            "rgfo_half_spin",
            "nso_half_spin",
            "sw_bias_gain_mode",
            "st_bias_gain_mode",
            "data_quality",
            "spin_period",
        ],
        "variable_names": LO_IALIRT_VARIABLE_NAMES,
    },
    CODICEAPID.COD_LO_INST_COUNTS_AGGREGATED: {
        "dataset_name": "imap_codice_l1a_lo-counters-aggregated",
        "input_dims": {"esa_step": 128, "spin_sector_pairs": 6},
        "instrument": "lo",
        "num_counters": len(
            LO_COUNTERS_AGGREGATED_VARIABLE_NAMES
        ),  # The number of counters depends on the number of active counters
        "output_dims": {"spin_sector_pairs": 6, "esa_step": 128},
        "support_variables": [
            "energy_table",
            "acquisition_time_per_step",
            "rgfo_half_spin",
            "nso_half_spin",
            "sw_bias_gain_mode",
            "st_bias_gain_mode",
            "data_quality",
            "spin_period",
        ],
        "variable_names": LO_COUNTERS_AGGREGATED_VARIABLE_NAMES,
    },
    CODICEAPID.COD_LO_INST_COUNTS_SINGLES: {
        "dataset_name": "imap_codice_l1a_lo-counters-singles",
        "input_dims": {"esa_step": 128, "inst_az": 24, "spin_sector_pairs": 6},
        "instrument": "lo",
        "num_counters": 1,
        "output_dims": {"inst_az": 24, "spin_sector_pairs": 6, "esa_step": 128},
        "support_variables": [
            "energy_table",
            "acquisition_time_per_step",
            "rgfo_half_spin",
            "nso_half_spin",
            "sw_bias_gain_mode",
            "st_bias_gain_mode",
            "data_quality",
            "spin_period",
        ],
        "variable_names": LO_COUNTERS_SINGLES_VARIABLE_NAMES,
    },
    CODICEAPID.COD_LO_SW_ANGULAR_COUNTS: {
        "dataset_name": "imap_codice_l1a_lo-sw-angular",
        "input_dims": {"esa_step": 128, "inst_az": 5, "spin_sector": 12},
        "instrument": "lo",
        "num_counters": 4,
        "output_dims": {"inst_az": 5, "spin_sector": 12, "esa_step": 128},
        "support_variables": [
            "energy_table",
            "acquisition_time_per_step",
            "rgfo_half_spin",
            "nso_half_spin",
            "sw_bias_gain_mode",
            "st_bias_gain_mode",
            "data_quality",
            "spin_period",
        ],
        "variable_names": LO_SW_ANGULAR_VARIABLE_NAMES,
    },
    CODICEAPID.COD_LO_NSW_ANGULAR_COUNTS: {
        "dataset_name": "imap_codice_l1a_lo-nsw-angular",
        "input_dims": {"esa_step": 128, "inst_az": 19, "spin_sector": 12},
        "instrument": "lo",
        "num_counters": 1,
        "output_dims": {"inst_az": 19, "spin_sector": 12, "esa_step": 128},
        "support_variables": [
            "energy_table",
            "acquisition_time_per_step",
            "rgfo_half_spin",
            "nso_half_spin",
            "sw_bias_gain_mode",
            "st_bias_gain_mode",
            "data_quality",
            "spin_period",
        ],
        "variable_names": LO_NSW_ANGULAR_VARIABLE_NAMES,
    },
    CODICEAPID.COD_LO_SW_PRIORITY_COUNTS: {
        "dataset_name": "imap_codice_l1a_lo-sw-priority",
        "input_dims": {"esa_step": 128, "spin_sector": 12},
        "instrument": "lo",
        "num_counters": 5,
        "output_dims": {"spin_sector": 12, "esa_step": 128},
        "support_variables": [
            "energy_table",
            "acquisition_time_per_step",
            "rgfo_half_spin",
            "nso_half_spin",
            "sw_bias_gain_mode",
            "st_bias_gain_mode",
            "data_quality",
            "spin_period",
        ],
        "variable_names": LO_SW_PRIORITY_VARIABLE_NAMES,
    },
    CODICEAPID.COD_LO_NSW_PRIORITY_COUNTS: {
        "dataset_name": "imap_codice_l1a_lo-nsw-priority",
        "input_dims": {"esa_step": 128, "spin_sector": 12},
        "instrument": "lo",
        "num_counters": 2,
        "output_dims": {"spin_sector": 12, "esa_step": 128},
        "support_variables": [
            "energy_table",
            "acquisition_time_per_step",
            "rgfo_half_spin",
            "nso_half_spin",
            "sw_bias_gain_mode",
            "st_bias_gain_mode",
            "data_quality",
            "spin_period",
        ],
        "variable_names": LO_NSW_PRIORITY_VARIABLE_NAMES,
    },
    CODICEAPID.COD_LO_SW_SPECIES_COUNTS: {
        "dataset_name": "imap_codice_l1a_lo-sw-species",
        "input_dims": {"esa_step": 128, "spin_sector": 1},
        "instrument": "lo",
        "num_counters": 16,
        "output_dims": {"spin_sector": 1, "esa_step": 128},
        "support_variables": [
            "energy_table",
            "acquisition_time_per_step",
            "rgfo_half_spin",
            "nso_half_spin",
            "sw_bias_gain_mode",
            "st_bias_gain_mode",
            "data_quality",
            "spin_period",
        ],
        "variable_names": LO_SW_SPECIES_VARIABLE_NAMES,
    },
    CODICEAPID.COD_LO_NSW_SPECIES_COUNTS: {
        "dataset_name": "imap_codice_l1a_lo-nsw-species",
        "input_dims": {"esa_step": 128, "spin_sector": 1},
        "instrument": "lo",
        "num_counters": 8,
        "output_dims": {"spin_sector": 1, "esa_step": 128},
        "support_variables": [
            "energy_table",
            "acquisition_time_per_step",
            "rgfo_half_spin",
            "nso_half_spin",
            "sw_bias_gain_mode",
            "st_bias_gain_mode",
            "data_quality",
            "spin_period",
        ],
        "variable_names": LO_NSW_SPECIES_VARIABLE_NAMES,
    },
}

# Various configurations to support L1b processing of individual data products
# Much of these are described in the algorithm document in chapter 11 ("Data
# Level 1B")
L1B_DATA_PRODUCT_CONFIGURATIONS: dict[str, dict] = {
    "hi-counters-aggregated": {
        "num_spin_sectors": 24,
        "num_spins": 16,
    },
    "hi-counters-singles": {
        "num_spin_sectors": 24,
        "num_spins": 16,
    },
    "hi-ialirt": {
        "num_spin_sectors": 24,
        "num_spins": 4,
    },
    "hi-omni": {
        "num_spin_sectors": 24,
        "num_spins": 4,
    },
    "hi-priority": {  # TODO: Ask Joey to define these
        "num_spin_sectors": 1,
        "num_spins": 1,
    },
    "hi-sectored": {
        "num_spin_sectors": 2,
        "num_spins": 16,
    },
    "lo-counters-aggregated": {
        "num_spin_sectors": 2,
    },
    "lo-counters-singles": {
        "num_spin_sectors": 2,
    },
    "lo-nsw-angular": {
        "num_spin_sectors": 1,
    },
    "lo-sw-angular": {
        "num_spin_sectors": 1,
    },
    "lo-nsw-priority": {
        "num_spin_sectors": 1,
    },
    "lo-sw-priority": {
        "num_spin_sectors": 1,
    },
    "lo-nsw-species": {
        "num_spin_sectors": 12,
    },
    "lo-sw-species": {
        "num_spin_sectors": 12,
    },
    "lo-ialirt": {
        "num_spin_sectors": 12,
    },
}

# Various configurations to support processing of direct events data products
# These are described in the algorithm document in chapter 10 ("Data Level 1A")
DE_DATA_PRODUCT_CONFIGURATIONS: dict[Any, dict[str, Any]] = {
    CODICEAPID.COD_HI_PHA: {
        "num_priorities": 6,
        "bit_structure": {
            "SSDEnergy": {
                "bit_length": 11,
                "dtype": np.uint16,
                "fillval": np.iinfo(np.uint16).max,
            },
            "TOF": {
                "bit_length": 10,
                "dtype": np.uint16,
                "fillval": np.iinfo(np.uint16).max,
            },
            "SSD_ID": {
                "bit_length": 4,
                "dtype": np.uint8,
                "fillval": np.iinfo(np.uint8).max,
            },
            "ERGE": {
                "bit_length": 2,
                "dtype": np.uint8,
                "fillval": np.iinfo(np.uint8).max,
            },
            "MultiFlag": {
                "bit_length": 1,
                "dtype": np.uint8,
                "fillval": np.iinfo(np.uint8).max,
            },
            "Type": {
                "bit_length": 2,
                "dtype": np.uint8,
                "fillval": np.iinfo(np.uint8).max,
            },
            "SpinAngle": {
                "bit_length": 5,
                "dtype": np.uint8,
                "fillval": np.iinfo(np.uint8).max,
            },
            "SpinNumber": {
                "bit_length": 4,
                "dtype": np.uint8,
                "fillval": np.iinfo(np.uint8).max,
            },
            "Priority": {
                "bit_length": 3,
                "dtype": np.uint8,
                "fillval": np.iinfo(np.uint8).max,
            },
            "Spare": {
                "bit_length": 22,
                "dtype": np.uint8,
                "fillval": np.iinfo(np.uint8).max,
            },
        },
        "cdf_fields": HI_PHA_CDF_FIELDS,
    },
    CODICEAPID.COD_LO_PHA: {
        "num_priorities": 8,
        "bit_structure": {
            "APDGain": {
                "bit_length": 1,
                "dtype": np.uint8,
                "fillval": np.iinfo(np.uint8).max,
            },
            "APD_ID": {
                "bit_length": 5,
                "dtype": np.uint8,
                "fillval": np.iinfo(np.uint8).max,
            },
            "Position": {
                "bit_length": 5,
                "dtype": np.uint8,
                "fillval": np.iinfo(np.uint8).max,
            },
            "APDEnergy": {
                "bit_length": 9,
                "dtype": np.uint16,
                "fillval": np.iinfo(np.uint16).max,
            },
            "TOF": {
                "bit_length": 10,
                "dtype": np.uint16,
                "fillval": np.iinfo(np.uint16).max,
            },
            "MultiFlag": {
                "bit_length": 1,
                "dtype": np.uint8,
                "fillval": np.iinfo(np.uint8).max,
            },
            "PHAType": {
                "bit_length": 2,
                "dtype": np.uint8,
                "fillval": np.iinfo(np.uint8).max,
            },
            "SpinAngle": {
                "bit_length": 5,
                "dtype": np.uint8,
                "fillval": np.iinfo(np.uint8).max,
            },
            "EnergyStep": {
                "bit_length": 7,
                "dtype": np.uint8,
                "fillval": np.iinfo(np.uint8).max,
            },
            "Priority": {
                "bit_length": 3,
                "dtype": np.uint8,
                "fillval": np.iinfo(np.uint8).max,
            },
            "Spare": {
                "bit_length": 16,
                "dtype": np.uint8,
                "fillval": np.iinfo(np.uint8).max,
            },
        },
        "cdf_fields": LO_PHA_CDF_FIELDS,
    },
}

# Define the packet fields needed to be stored in segmented data and their
# corresponding bit lengths for direct event data products
IAL_BIT_STRUCTURE = {
    "SHCOARSE": 32,
    "PACKET_VERSION": 16,
    "SPIN_PERIOD": 16,
    "ACQ_START_SECONDS": 32,
    "ACQ_START_SUBSECONDS": 20,
    "SPARE_00": 8,
    "ST_BIAS_GAIN_MODE": 2,
    "SW_BIAS_GAIN_MODE": 2,
    "TABLE_ID": 32,
    "PLAN_ID": 16,
    "PLAN_STEP": 4,
    "VIEW_ID": 4,
    "RGFO_HALF_SPIN": 6,
    "NSO_HALF_SPIN": 6,
    "SPARE_01": 1,
    "SUSPECT": 1,
    "COMPRESSION": 3,
    "BYTE_COUNT": 23,
}

# Define the packet fields needed to be stored in segmented data and their
# corresponding bit lengths for direct event data products
DE_METADATA_FIELDS = {
    "packet_version": 16,
    "spin_period": 16,
    "acq_start_seconds": 32,
    "acq_start_subseconds": 20,
    "spare_1": 2,
    "st_bias_gain_mode": 2,
    "sw_bias_gain_mode": 2,
    "priority": 4,
    "suspect": 1,
    "compressed": 1,
    "num_events": 32,
    "byte_count": 32,
}

# Compression ID lookup tables
# The key is the view_id and the value is the ID for the compression algorithm
# (see utils.CoDICECompression to see how the values correspond)
# These are defined in the "Views" tab of the "*-SCI-LUT-*.xml" spreadsheet that
# largely defines CoDICE processing.
LO_COMPRESSION_ID_LOOKUP = {
    0: CoDICECompression.PACK_24_BIT,
    1: CoDICECompression.LOSSY_B_LOSSLESS,
    2: CoDICECompression.LOSSY_B_LOSSLESS,
    3: CoDICECompression.LOSSY_A_LOSSLESS,
    4: CoDICECompression.LOSSY_A_LOSSLESS,
    5: CoDICECompression.LOSSY_A_LOSSLESS,
    6: CoDICECompression.LOSSY_A_LOSSLESS,
    7: CoDICECompression.LOSSY_A_LOSSLESS,
    8: CoDICECompression.LOSSY_A_LOSSLESS,
}
HI_COMPRESSION_ID_LOOKUP = {
    0: CoDICECompression.LOSSY_A,
    1: CoDICECompression.LOSSY_A,
    2: CoDICECompression.LOSSY_A,
    3: CoDICECompression.LOSSY_B_LOSSLESS,
    4: CoDICECompression.LOSSY_B_LOSSLESS,
    5: CoDICECompression.LOSSY_A_LOSSLESS,
    6: CoDICECompression.LOSSY_A_LOSSLESS,
    7: CoDICECompression.LOSSY_A_LOSSLESS,
    8: CoDICECompression.LOSSY_A_LOSSLESS,
    9: CoDICECompression.LOSSY_A_LOSSLESS,
}

# ESA Sweep table ID lookup table
# The combination of plan_id and plan_step determine the ESA sweep Table to use
# Currently, ESA sweep table 0 is used for every plan_id/plan_step combination,
# but may change in the future. These are defined in the "ESA Sweep" tab of the
# "*-SCI-LUT-*.xml" spreadsheet that largely defines CoDICE processing.
ESA_SWEEP_TABLE_ID_LOOKUP = {
    (0, 0): 0,
    (0, 1): 0,
    (0, 2): 0,
    (0, 3): 0,
    (1, 0): 0,
    (1, 1): 0,
    (1, 2): 0,
    (1, 3): 0,
    (2, 0): 0,
    (2, 1): 0,
    (2, 2): 0,
    (2, 3): 0,
    (3, 0): 0,
    (3, 1): 0,
    (3, 2): 0,
    (3, 3): 0,
    (4, 0): 0,
    (4, 1): 0,
    (4, 2): 0,
    (4, 3): 0,
    (5, 0): 0,
    (5, 1): 0,
    (5, 2): 0,
    (5, 3): 0,
    (6, 0): 0,
    (6, 1): 0,
    (6, 2): 0,
    (6, 3): 0,
    (7, 0): 0,
    (7, 1): 0,
    (7, 2): 0,
    (7, 3): 0,
}

# Lo Stepping table ID lookup table
# The combination of plan_id and plan_step determine the Lo Stepping Table to
# use. Currently, LO Stepping table 0 is used for every plan_id/plan_step
# combination, but may change in the future. These are defined in the "Lo
# Stepping" tab of the "*-SCI-LUT-*.xml" spreadsheet that largely defines CoDICE
# processing.
LO_STEPPING_TABLE_ID_LOOKUP = {
    (0, 0): 0,
    (0, 1): 0,
    (0, 2): 0,
    (0, 3): 0,
    (1, 0): 0,
    (1, 1): 0,
    (1, 2): 0,
    (1, 3): 0,
    (2, 0): 0,
    (2, 1): 0,
    (2, 2): 0,
    (2, 3): 0,
    (3, 0): 0,
    (3, 1): 0,
    (3, 2): 0,
    (3, 3): 0,
    (4, 0): 0,
    (4, 1): 0,
    (4, 2): 0,
    (4, 3): 0,
    (5, 0): 0,
    (5, 1): 0,
    (5, 2): 0,
    (5, 3): 0,
    (6, 0): 0,
    (6, 1): 0,
    (6, 2): 0,
    (6, 3): 0,
    (7, 0): 0,
    (7, 1): 0,
    (7, 2): 0,
    (7, 3): 0,
}

# Lookup tables for Lossy decompression algorithms "A" and "B"
# These were provided by Greg Dunn via his sohis_cdh_utils.v script and then
# transformed into python dictionaries. The values in these tables are subject
# to change, but the format is expected to stay the same.
LOSSY_A_TABLE = {
    0: 0,
    1: 1,
    2: 2,
    3: 3,
    4: 4,
    5: 5,
    6: 6,
    7: 7,
    8: 8,
    9: 9,
    10: 10,
    11: 11,
    12: 12,
    13: 13,
    14: 14,
    15: 15,
    16: 16,
    17: 17,
    18: 18,
    19: 19,
    20: 20,
    21: 21,
    22: 22,
    23: 23,
    24: 24,
    25: 25,
    26: 26,
    27: 27,
    28: 28,
    29: 29,
    30: 30,
    31: 31,
    32: 33,
    33: 35,
    34: 37,
    35: 39,
    36: 41,
    37: 43,
    38: 45,
    39: 47,
    40: 49,
    41: 51,
    42: 53,
    43: 55,
    44: 57,
    45: 59,
    46: 61,
    47: 63,
    48: 67,
    49: 71,
    50: 75,
    51: 79,
    52: 83,
    53: 87,
    54: 91,
    55: 95,
    56: 99,
    57: 103,
    58: 107,
    59: 111,
    60: 115,
    61: 119,
    62: 123,
    63: 127,
    64: 135,
    65: 143,
    66: 151,
    67: 159,
    68: 167,
    69: 175,
    70: 183,
    71: 191,
    72: 199,
    73: 207,
    74: 215,
    75: 223,
    76: 231,
    77: 239,
    78: 247,
    79: 255,
    80: 271,
    81: 287,
    82: 303,
    83: 319,
    84: 335,
    85: 351,
    86: 367,
    87: 383,
    88: 399,
    89: 415,
    90: 431,
    91: 447,
    92: 463,
    93: 479,
    94: 495,
    95: 511,
    96: 543,
    97: 575,
    98: 607,
    99: 639,
    100: 671,
    101: 703,
    102: 735,
    103: 767,
    104: 799,
    105: 831,
    106: 863,
    107: 895,
    108: 927,
    109: 959,
    110: 991,
    111: 1023,
    112: 1087,
    113: 1151,
    114: 1215,
    115: 1279,
    116: 1343,
    117: 1407,
    118: 1471,
    119: 1535,
    120: 1599,
    121: 1663,
    122: 1727,
    123: 1791,
    124: 1855,
    125: 1919,
    126: 1983,
    127: 2047,
    128: 2175,
    129: 2303,
    130: 2431,
    131: 2559,
    132: 2687,
    133: 2815,
    134: 2943,
    135: 3071,
    136: 3199,
    137: 3327,
    138: 3455,
    139: 3583,
    140: 3711,
    141: 3839,
    142: 3967,
    143: 4095,
    144: 4351,
    145: 4607,
    146: 4863,
    147: 5119,
    148: 5375,
    149: 5631,
    150: 5887,
    151: 6143,
    152: 6399,
    153: 6655,
    154: 6911,
    155: 7167,
    156: 7423,
    157: 7679,
    158: 7935,
    159: 8191,
    160: 8703,
    161: 9215,
    162: 9727,
    163: 10239,
    164: 10751,
    165: 11263,
    166: 11775,
    167: 12287,
    168: 12799,
    169: 13311,
    170: 13823,
    171: 14335,
    172: 14847,
    173: 15359,
    174: 15871,
    175: 16383,
    176: 17407,
    177: 18431,
    178: 19455,
    179: 20479,
    180: 21503,
    181: 22527,
    182: 23551,
    183: 24575,
    184: 25599,
    185: 26623,
    186: 27647,
    187: 28671,
    188: 29695,
    189: 30719,
    190: 31743,
    191: 32767,
    192: 34815,
    193: 36863,
    194: 38911,
    195: 40959,
    196: 43007,
    197: 45055,
    198: 47103,
    199: 49151,
    200: 51199,
    201: 53247,
    202: 55295,
    203: 57343,
    204: 59391,
    205: 61439,
    206: 63487,
    207: 65535,
    208: 69631,
    209: 73727,
    210: 77823,
    211: 81919,
    212: 86015,
    213: 90111,
    214: 94207,
    215: 98303,
    216: 102399,
    217: 106495,
    218: 110591,
    219: 114687,
    220: 118783,
    221: 122879,
    222: 126975,
    223: 131071,
    224: 139263,
    225: 147455,
    226: 155647,
    227: 163839,
    228: 172031,
    229: 180223,
    230: 188415,
    231: 196607,
    232: 204799,
    233: 212991,
    234: 221183,
    235: 229375,
    236: 237567,
    237: 245759,
    238: 253951,
    239: 262143,
    240: 278527,
    241: 294911,
    242: 311295,
    243: 327679,
    244: 344063,
    245: 360447,
    246: 376831,
    247: 393215,
    248: 409599,
    249: 425983,
    250: 442367,
    251: 458751,
    252: 475135,
    253: 491519,
    254: 507903,
    255: 4294967294,
}

LOSSY_B_TABLE = {
    0: 0,
    1: 1,
    2: 2,
    3: 3,
    4: 4,
    5: 5,
    6: 6,
    7: 7,
    8: 8,
    9: 9,
    10: 10,
    11: 11,
    12: 12,
    13: 13,
    14: 14,
    15: 15,
    16: 16,
    17: 17,
    18: 18,
    19: 19,
    20: 20,
    21: 21,
    22: 22,
    23: 23,
    24: 24,
    25: 25,
    26: 26,
    27: 27,
    28: 28,
    29: 29,
    30: 30,
    31: 31,
    32: 33,
    33: 35,
    34: 37,
    35: 39,
    36: 41,
    37: 43,
    38: 45,
    39: 47,
    40: 49,
    41: 51,
    42: 53,
    43: 55,
    44: 57,
    45: 59,
    46: 61,
    47: 63,
    48: 67,
    49: 71,
    50: 75,
    51: 79,
    52: 83,
    53: 87,
    54: 91,
    55: 95,
    56: 99,
    57: 103,
    58: 107,
    59: 111,
    60: 115,
    61: 119,
    62: 123,
    63: 127,
    64: 135,
    65: 143,
    66: 151,
    67: 159,
    68: 167,
    69: 175,
    70: 183,
    71: 191,
    72: 199,
    73: 207,
    74: 215,
    75: 223,
    76: 231,
    77: 239,
    78: 247,
    79: 255,
    80: 271,
    81: 287,
    82: 303,
    83: 319,
    84: 335,
    85: 351,
    86: 367,
    87: 383,
    88: 399,
    89: 415,
    90: 431,
    91: 447,
    92: 463,
    93: 479,
    94: 495,
    95: 511,
    96: 543,
    97: 575,
    98: 607,
    99: 639,
    100: 671,
    101: 703,
    102: 735,
    103: 767,
    104: 799,
    105: 831,
    106: 863,
    107: 895,
    108: 927,
    109: 959,
    110: 991,
    111: 1023,
    112: 1087,
    113: 1151,
    114: 1215,
    115: 1279,
    116: 1343,
    117: 1407,
    118: 1471,
    119: 1535,
    120: 1599,
    121: 1663,
    122: 1727,
    123: 1791,
    124: 1855,
    125: 1919,
    126: 1983,
    127: 2047,
    128: 2175,
    129: 2303,
    130: 2431,
    131: 2559,
    132: 2687,
    133: 2815,
    134: 2943,
    135: 3071,
    136: 3199,
    137: 3327,
    138: 3455,
    139: 3583,
    140: 3711,
    141: 3839,
    142: 3967,
    143: 4095,
    144: 4351,
    145: 4607,
    146: 4863,
    147: 5119,
    148: 5375,
    149: 5631,
    150: 5887,
    151: 6143,
    152: 6399,
    153: 6655,
    154: 6911,
    155: 7167,
    156: 7423,
    157: 7679,
    158: 7935,
    159: 8191,
    160: 8703,
    161: 9215,
    162: 9727,
    163: 10239,
    164: 10751,
    165: 11263,
    166: 11775,
    167: 12287,
    168: 12799,
    169: 13311,
    170: 13823,
    171: 14335,
    172: 14847,
    173: 15359,
    174: 15871,
    175: 16383,
    176: 17407,
    177: 18431,
    178: 19455,
    179: 20479,
    180: 21503,
    181: 22527,
    182: 23551,
    183: 24575,
    184: 25599,
    185: 26623,
    186: 27647,
    187: 28671,
    188: 29695,
    189: 30719,
    190: 31743,
    191: 32767,
    192: 36863,
    193: 40959,
    194: 45055,
    195: 49151,
    196: 53247,
    197: 57343,
    198: 61439,
    199: 65535,
    200: 73727,
    201: 81919,
    202: 90111,
    203: 98303,
    204: 106495,
    205: 114687,
    206: 122879,
    207: 131071,
    208: 147455,
    209: 163839,
    210: 180223,
    211: 196607,
    212: 212991,
    213: 229375,
    214: 245759,
    215: 262143,
    216: 294911,
    217: 327679,
    218: 360447,
    219: 393215,
    220: 425983,
    221: 458751,
    222: 491519,
    223: 524287,
    224: 589823,
    225: 655359,
    226: 720895,
    227: 786431,
    228: 851967,
    229: 917503,
    230: 983039,
    231: 1048575,
    232: 1179647,
    233: 1310719,
    234: 1441791,
    235: 1572863,
    236: 1703935,
    237: 1835007,
    238: 1966079,
    239: 2097151,
    240: 2359295,
    241: 2621439,
    242: 2883583,
    243: 3145727,
    244: 3407871,
    245: 3670015,
    246: 3932159,
    247: 4194303,
    248: 4718591,
    249: 5242879,
    250: 5767167,
    251: 6291455,
    252: 6815743,
    253: 7340031,
    254: 7864319,
    255: 4294967294,
}

# Derived acquisition times that get stored in CDF data variables in L1a
# processing. These are taken from the "Acq Time" column in the "Lo Stepping"
# tab of the "*-SCI-LUT-*.xml" spreadsheet that largely defines CoDICE
# processing.
ACQUISITION_TIMES = {
    0: [
        578.7083,
        578.7083,
        578.7083,
        578.7083,
        289.35416,
        289.35416,
        289.35416,
        289.35416,
        289.35416,
        289.35416,
        289.35416,
        289.35416,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
    ],
    1: [
        578.7083,
        578.7083,
        578.7083,
        578.7083,
        289.35416,
        289.35416,
        289.35416,
        289.35416,
        289.35416,
        289.35416,
        289.35416,
        289.35416,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
    ],
    2: [
        578.7083,
        578.7083,
        578.7083,
        578.7083,
        289.35416,
        289.35416,
        289.35416,
        289.35416,
        289.35416,
        289.35416,
        289.35416,
        289.35416,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
    ],
    3: [
        578.7083,
        578.7083,
        578.7083,
        578.7083,
        289.35416,
        289.35416,
        289.35416,
        289.35416,
        289.35416,
        289.35416,
        289.35416,
        289.35416,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        192.90277,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        144.67708,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        115.74167,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
        95.69444,
    ],
}
